"""
Core encoding/decoding logic of the library
"""

from io import TextIOBase, StringIO
from typing import Optional

from .handlers import BaudotReader, BaudotWriter
from .codecs import BaudotCodec, Shift


def encode(stream: TextIOBase, codec: BaudotCodec, writer: BaudotWriter):
    """
    Encode unicode characters from an input stream to an output writer,
    using the given codec.

    :param stream: Unicode character stream to encode (can be a file)
    :param codec: Codec to use for encoding
    :param writer: Writer instance for the wanted output format
    """
    state: Optional[Shift] = None
    buffer = []

    while True:
        char = stream.read(1)
        if not char:  # TextIOBase returns empty character on EOF
            break

        code, new_state = codec.encode(char, state)
        buffer.append(code)

        if new_state != state:
            state_code, _ = codec.encode(new_state, None)
            buffer.append(state_code)
            state = new_state

        while buffer:
            writer.write(buffer.pop(-1))


def encode_str(chars: str, codec: BaudotCodec, writer: BaudotWriter):
    """
    Encode unicode characters from an input string to an output writer,
    using the given codec.

    :param chars: Unicode string to encode
    :param codec: Codec to use for encoding
    :param writer: Writer instance for the wanted output format
    """
    with StringIO(chars) as stream:
        encode(stream, codec, writer)


def decode(reader: BaudotReader, codec: BaudotCodec, stream: TextIOBase):
    """
    Decode a baudot code stream from a reader to a unicode stream,
    using a given codec.

    :param reader: Reader instance that will read codes from an input
    :param codec: Codec to use for decoding
    :param stream: Unicode stream to write to (can be a file)
    """
    state: Optional[Shift] = None

    for code in reader:
        value = codec.decode(code, state)

        if isinstance(value, Shift):
            state = value
        else:
            stream.write(value)


def decode_to_str(reader: BaudotReader, codec: BaudotCodec) -> str:
    """
    Decode a baudot code stream from a reader to a unicode string,
    using a given codec.

    :param reader: Reader instance that will read codes from an input
    :param codec: Codec to use for decoding
    :return: Decoded Unicode string
    """
    with StringIO('') as stream:
        decode(reader, codec, stream)
        return stream.getvalue()




"""
Core codec base classes and logic.

These tools handle encoding and decoding *per character*;
the core logic of stateful encoding is in ``baudot.core``
"""

from abc import ABC, abstractmethod
from collections import namedtuple, defaultdict
from typing import List, Dict, Union, Tuple, Set, Optional

from ..exceptions import IncoherentTable, DecodingError, EncodingError

__all__ = ['Shift', 'BaudotCodec', 'SimpleTabledCodec']

Shift = namedtuple('Shift', ('name',))

Value = Union[str, Shift]
Table = List[Value]


class BaudotCodec(ABC):
    """
    Abstract Base Class for a Codec

    Subclasses must implement :py:meth:`encode` and :py:meth:`decode`
    """

    @abstractmethod
    def encode(self, value: Value, state: Shift) -> Tuple[int, Shift]:
        """
        Abstract method for encoding a single character or state shift
        """

    @abstractmethod
    def decode(self, code: int, state: Shift) -> Union[str, Shift]:
        """
        Abstract method for decoding a single code.
        """


class SimpleTabledCodec(BaudotCodec):
    """
    Creates a codec based on a character table.

    The input format must be a dictionary of which the keys are the
    possible states (instances of ``Shift``) and the values are lists
    of length 32 exactly, containing characters or shifts.

    The ``Shift`` instances are the only control characters this
    library knows of. Any other must be taken from ASCII/Unicode.
    """

    __slots__ = ['name', 'shifts', 'alphabet', 'decoding_table',
                 'encoding_single', 'encoding_any', 'encoding_others']

    def __init__(self, name: str, tables: Dict[Shift, Table]):

        _verify_tables(tables)
        enc_single, enc_any, enc_others = _make_simple_encoding_table(tables)

        self.name: str = name
        self.shifts: Set[Shift] = set(tables.keys())
        self.alphabet: Set[str] = set(value
                                      for table in tables.values()
                                      for value in table
                                      if isinstance(value, str))
        self.decoding_table: Dict[Shift, Table] = tables
        self.encoding_single: Dict[Value, Tuple[int, Shift]] = enc_single
        self.encoding_any: Dict[Value, int] = enc_any
        self.encoding_others: Dict[Value, Set[Tuple[int, Shift]]] = enc_others

    def encode(self, value: Value, state: Shift) -> Tuple[int, Shift]:
        """
        Get the code of the given character of Shift for this codec.

        Actually, this logic returns not only the code but also the
        state required for this code. The current state should also
        be passed so that more complicated cases can be solved.

        :param value: Value (character or state shift) to encode
        :param state: Current state of encoding
        :return: Code for this value, and required state
        """

        # If this value exists (with the same code) in all states,
        # return its code and the state unchanged
        if value in self.encoding_any:
            return self.encoding_any[value], state

        # If the value unambiguously exists in only one state,
        # return its code and the corresponding state
        if value in self.encoding_single:
            return self.encoding_single[value]

        if value not in self.encoding_others:
            raise EncodingError(f"Unsupported value {value}")

        # The logic below handles other cases. This will need more work.

        # TODO: Allow char with codes in two states but not in a third
        # TODO: Allow char that appears twice in a single state
        matches = self.encoding_others[value]
        try:
            # This logic handles the case where a same character exists
            # in multiple states but with different codes. Just return
            # the code that works with the current state.
            return next((code, st) for code, st in matches if st == state)
        except StopIteration:
            raise NotImplementedError("This char encodes in a weird way.")

    def decode(self, code: int, state: Optional[Shift]) -> Value:
        """
        Get the character or state shift corresponding to a given
        code in a given state.

        :param code: Code to look up
        :param state: State to apply. This may be `None`, so that a
            the state can be initialized.
        :return: Decoded character or state shift
        """

        if not 0 <= code < 32:
            raise DecodingError(f"Invalid code: {code}")

        # Allow state initialization by looking up unambiguous values
        if state is None and code in self.encoding_any.values():
            return next(val for val, c in self.encoding_any.items() if c == code)

        if state not in self.decoding_table:
            raise DecodingError(f"Unrecognized state: {state}")

        return self.decoding_table[state][code]


def _verify_tables(tables: Dict[Shift, Table]):
    """
    Function for verifying that a given input table is correct
    """

    if not tables:
        raise IncoherentTable('A coding table cannot be empty')

    if not all(len(t) == 32 for t in tables.values()):
        raise IncoherentTable('All tables MUST have 32 entries')

    shifts_in_keys = set(tables.keys())
    shifts_in_table = set(shift
                          for table in tables.values()
                          for shift in table
                          if isinstance(shift, Shift))

    if shifts_in_keys != shifts_in_table:
        raise IncoherentTable("Shifts in the tables don't match their keys")


def _make_simple_encoding_table(tables: Dict[Shift, Table]):
    """
    Generates the encoding tables by reversing the decoding table
    """

    matches: Dict[Value, Set[Tuple[int, Shift]]] = defaultdict(set)
    for shift, table in tables.items():
        for i, value in enumerate(table):
            matches[value].add((i, shift))

    single_match, any_matches, others = {}, {}, {}

    for value, match in matches.items():

        if len(match) == 1:
            single_match[value] = match.pop()
            continue

        codes = set(code for code, _ in match)
        if len(codes) == 1 and len(match) == len(tables):
            any_matches[value] = codes.pop()
            continue

        others[value] = match

    return single_match, any_matches, others




"""
Codecs are the tools used to convert encoded-data (5-bit digits)
into Unicode characters and back.
"""

from .core import BaudotCodec, SimpleTabledCodec, Shift
from .ita1_baudot import ITA1_CONTINENTAL, ITA1_UK
from .ita2_baudot_murray import ITA2_STANDARD, ITA2_US




"""
The handlers are interfaces to read and write 5-bit data
from a variety of formats.
"""

from .core import BaudotWriter, BaudotReader
from .tape import TapeReader, TapeWriter, TapeConfig
from .hexbytes import HexBytesReader, HexBytesWriter



"""
Codec definitions for second-generation Baudot codes
(a.k.a. Baudot-Murray or ITA2)
"""
# pylint: disable=invalid-name

from .core import Shift, SimpleTabledCodec

Letters = Shift('Letters')
Figures = Shift('Figures')

STANDARD_TABLE = {
    Letters: [
        '\x00', 'E', '\n', 'A', ' ', 'S', 'I', 'U',
        '\r', 'D', 'R', 'J', 'N', 'F', 'C', 'K',
        'T', 'Z', 'L', 'W', 'H', 'Y', 'P', 'Q',
        'O', 'B', 'G', Figures, 'M', 'X', 'V', Letters
    ],
    Figures: [
        '\x00', '3', '\n', '-', ' ', "'", '8', '7',
        '\r', '\x05', '4', '\x07', ',', '!', ':', '(',
        '5', '+', ')', '2', '£', '6', '0', '1',
        '9', '?', '&', Figures, '.', '/', '=', Letters
    ],
}

US_TABLE = {
    Letters: STANDARD_TABLE[Letters].copy(),
    Figures: [
        '\x00', '3', '\n', '-', ' ', '\x07', '8', '7',
        '\r', '$', '4', "'", ',', '!', ':', '(',
        '5', '"', ')', '2', '#', '6', '0', '1',
        '9', '?', '&', Figures, '.', '/', ';', Letters
    ],
}

ITA2_STANDARD = SimpleTabledCodec(
    'Baudot-Murray code (ITA2), Standard',
    STANDARD_TABLE)
ITA2_US = SimpleTabledCodec(
    'Baudot-Murray code (ITA2), US variant',
    US_TABLE)



"""
Core utilities for Baudot input/output handlers
"""

from abc import ABC, abstractmethod


class BaudotReader(ABC):
    """Abstract Base Class for a reader"""

    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self):
        pass


class BaudotWriter(ABC):
    """Abstract Base Class for a writer"""

    @abstractmethod
    def write(self, code: int):
        """Write a single code to the output"""



"""
Handler for reading and writing 5-bit codes as a hexadecimal bit stream.
"""

from io import BufferedIOBase

from .core import BaudotReader, BaudotWriter
from ..exceptions import ReadError, WriteError


class HexBytesReader(BaudotReader):
    """
    Reader for hexadecimal 5-bit streams
    """

    def __init__(self, stream: BufferedIOBase):
        self.stream = stream

    def __next__(self):
        hex_byte = self.stream.read(2)
        if not hex_byte:
            raise StopIteration()
        try:
            code = int(hex_byte, 16)
        except ValueError:
            str_repr = hex_byte.decode(errors='backslashreplace')
            raise ReadError(f'Invalid hexadecimal byte: {str_repr}')
        if not 0 <= code < 32:
            raise ReadError(f'Code value {code} is not a valid 5-bit value')
        return code


class HexBytesWriter(BaudotWriter):
    """
    Writer for hexadecimal 5-bit stream
    """

    def __init__(self, stream: BufferedIOBase):
        self.stream = stream

    def write(self, code: int):
        """Writes a code as an hexadecimal value"""
        if not 0 <= code < 32:
            raise WriteError('Invalid 5-bit character code')

        self.stream.write(f'{code:02x}'.encode())




"""
Custom exceptions for the Baudot library
"""


class BaudotException(Exception):
    """Core exception class for this library"""


class IncoherentTable(BaudotException):
    """Raised when an encoding/decoding table is not valid"""


class DecodingError(BaudotException):
    """Raised on decoding error"""


class EncodingError(BaudotException):
    """Raised on encoding error"""


class ReadError(BaudotException):
    """Raised when reading a 5-bit stream fails"""


class WriteError(BaudotException):
    """Raised when writing a 5-bit stream fails"""




from setuptools import setup, find_packages

LONG_DESC = """
Baudot is a Python library for encoding and decoding 5-bit stateful encoding.

This library is named after `Jean-Maurice-Émile Baudot`_ (1845-1903), the French
engineer who invented this code. The `Baudot code`_ was the first practical and
widely used binary character encoding, and is an ancestor of the ASCII code we
are familiar with today.

This library accomplishes two tasks (and their inverse):

1. reading 5-bit codes from custom input formats,
2. converting 5-bit codes to unicode characters.

This library provides support for several common encodings, as well as support
for reading tape-like files and hexadecimal data.

Please keep in mind that this project is very young, and that its API is most
likely ill-designed at this point. Suggestions are welcome!

.. _`Jean-Maurice-Émile Baudot`: https://en.wikipedia.org/wiki/%C3%89mile_Baudot
.. _`Baudot code`: https://en.wikipedia.org/wiki/Baudot_code

"""

setup(
    name='baudot',
    version='0.1.1.post2',
    description='Tools for handling stateful 5-bit encoding',
    long_description=LONG_DESC,
    author='Xavier Villaneau',
    author_email='xvillaneau+baudot@gmail.com',
    url='https://github.com/xvillaneau/python-baudot',
    license='LGPLv3',
    python_requires='>=3.6.0',
    packages=find_packages(exclude=('tests',)),
    test_requires=["pytest", "hypothesis", "coverage", "pylint"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications',
    ],
)



"""
Handler for reading and writing to pretty tape-like formatted text

For example, the tape might look like:::

    ***.**
    * *.
       . *
    *  .*
    *  .*
    ** .
      *.
    *  .**
    ** .
     * .*
    *  .*
     * . *
    ** .**
     **. *

(Which reads 'HELLO WORLD!')
"""

from collections import namedtuple
from io import TextIOBase

from .core import BaudotReader, BaudotWriter
from ..exceptions import WriteError

TapeConfig = namedtuple('TapeConfig', ('punch', 'blank', 'sep'))
TapeConfig.__doc__ = """
Object for storing a tape representation format.
"""

DEFAULT_TAPE = TapeConfig('*', ' ', '.')

MSB_FIRST = [1 << (4-n) for n in range(5)]


class TapeReader(BaudotReader):
    """
    Reader class for tape-like data.
    """

    def __init__(self, stream: TextIOBase, config: TapeConfig = DEFAULT_TAPE):
        self.stream = stream
        self.config = config

    def __next__(self) -> int:
        line = next(self.stream)
        pairs = zip(line.replace(self.config.sep, ''), MSB_FIRST)
        return sum(n for c, n in pairs if c == self.config.punch)


class TapeWriter(BaudotWriter):
    """
    Writer class for tape-like data.
    """

    def __init__(self, stream: TextIOBase, config: TapeConfig = DEFAULT_TAPE):
        self.stream = stream
        self.config = config

    def write(self, code: int):
        """Writes a code to tape"""
        if not 0 <= code < 32:
            raise WriteError('Invalid 5-bit character code')

        chars = ''.join(self.config.punch if c == '1' else self.config.blank
                        for c in f'{code:05b}')
        self.stream.write(f"{chars[:3]}{self.config.sep}{chars[3:]}\n")
